#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "sorter.h"

enum read_state {
    LEN = 0,
    ARR1 = 1,
    ARR2 = 2,
    FINISH = 3,
    IMPOSSIBLE_STATE = 4
};

static int to_int(char *a)
{
    int value = 0;

    for (int i = 0; i < 4; i++) {
        if (a[i] == '\0')
            break;

        value *= 10;
        value += (int) a[i] - (int) '0';
    }

    return value;
}

static void inplace_memcpy(char *buf, int offset)
{
    for (int i = 0; i < 1024 - offset; i++)
        buf[i] = buf[offset+i];
}

static int read_more(int f, char *buf, int bytes_read)
{
    // in-place memcpy
    inplace_memcpy(buf, bytes_read);

    // reading more data
    int new_bytes = read(f, &(buf[1024 - bytes_read]), bytes_read);

    // returning current amount of bytes in buffer
    return 1024 - bytes_read + new_bytes;
}

static int parse_case(char *cmd, int *array_size, char ***a, char ***b)
{
    int bytes_read;
    int write_offset = 0;
    int array_offset = 0;
    char len_buffer[4] = {0};
    enum read_state rs = LEN;

    for (bytes_read = 0; bytes_read < 1024; bytes_read++) {
        switch (cmd[bytes_read]) {
        case ',':
            switch (rs) {
            case ARR1:
                (*a)[array_offset][write_offset] = '\0';
                break;
            case ARR2:
                (*b)[array_offset][write_offset] = '\0';
                break;
            }

            write_offset = 0;
            array_offset++;
            break;

        case ' ':
            switch (rs) {
            case LEN:
                *array_size = to_int(len_buffer);

                *a = (char **) malloc(sizeof(char *) * (*array_size));
                *b = (char **) malloc(sizeof(char *) * (*array_size));

                for (int i = 0; i < *array_size; i++) {
                    (*a)[i] = malloc(sizeof(char) * 128);
                    (*b)[i] = malloc(sizeof(char) * 128);
                }

                break;
            case ARR1:
                (*a)[array_offset][write_offset] = '\0';
                break;
            }
            rs++;
            write_offset = 0;
            array_offset = 0;
            break;

        case '\n':
            (*b)[array_offset][write_offset] = '\0';
            bytes_read++;
            goto finish;
            break;

        default:
            switch(rs) {
            case LEN:
                len_buffer[write_offset] = cmd[bytes_read];
                break;
            case ARR1:
                (*a)[array_offset][write_offset] = cmd[bytes_read];
                break;
            case ARR2:
                (*b)[array_offset][write_offset] = cmd[bytes_read];
                break;
            }
            write_offset++;
            break;
        }
    }

    finish:
    return bytes_read;
}

static void destroy_array(char **a, int array_size)
{
    for (int i = 0; i < array_size; i++)
        free(a[i]);

    free(a);
}

static int cmp_arrays(int array_size, char **a, char **b)
{
    int mismatch_byte = 0;

    for (int i = 0; i < array_size; i++, mismatch_byte++) {
        int la = strlen(a[i]);
        int lb = strlen(b[i]);

        if (la != lb)
            return mismatch_byte;

        for (int j = 0; j < la; j++, mismatch_byte++) {
            if (a[i][j] != b[i][j])
                return mismatch_byte;
        }
    }

    return -1;
}

static void print_array(int array_size, char **arr)
{
    for (int i = 0; i < array_size; i++)
        printf("%s ", arr[i]);

    puts("");
}

static void print_mismatch(int position)
{
    for (int i = 0; i < position; i++)
        fputs(" ", stdout);

    puts("^");

    for (int i = 0; i < position; i++)
        fputs(" ", stdout);

    puts("sort mismatch");
}

static int assert_sort(int test_id, int array_size, char **a, char **b)
{
    sorter_insertion_sort(a, array_size);

    int result = cmp_arrays(array_size, a, b);

    if (result == -1) {
        printf("[TEST#%02d] OK\r", test_id);
        return 0;
    } else {
        printf("\n[TEST#%02d] FAILED!\n", test_id);
        printf("Case details:\n");
        printf("Retrieved array: ");
        print_array(array_size, a);

        printf("Expected  array: ");
        print_array(array_size, b);

        print_mismatch(result + 17);

        return 1;
    }
}

int main(int argc, char **argv)
{
    int success_tests_counter = 0;
    int failed_tests_counter = 0;

    int f;

    if (argc == 2)
        f = open(argv[1], O_RDONLY);
    else
        f = open("test/testcases_sort_official", O_RDONLY);

    if (f < 0)
        return 1;

    char cmd[1024];
    int remaining_buffer_length;
    remaining_buffer_length = read(f, cmd, 1024);

    while (remaining_buffer_length) {
        char **a, **b;
        int array_size;

        int bytes_read = parse_case(cmd, &array_size, &a, &b);
        
        if (assert_sort(success_tests_counter + failed_tests_counter + 1, array_size, a, b))
            failed_tests_counter++;
        else
            success_tests_counter++;

        // clean up before next iteration
        destroy_array(a, array_size);
        destroy_array(b, array_size);

        if (remaining_buffer_length == 1024)
            remaining_buffer_length = read_more(f, cmd, bytes_read);
        else
            inplace_memcpy(cmd, bytes_read);
            remaining_buffer_length -= bytes_read;
    }

    printf("\nSummary: %d passed, %d failed\n", success_tests_counter, failed_tests_counter);
    close(f);

    return !(!failed_tests_counter);
}
